/**
 * @file i2c_test.c
 *
 * This file provides an example test interface for the reading/writing
 * to the Linux I2C device interface. This code is only relevant if
 * the I2C Device interface is enabled and an external I2C device is
 * connected to the bus. I2C is not enabled by default on the SoM-9260M
 * but is available as a module (using any two available PIO pins on the
 * SOM). The SoM-9G20M includes I2C support by default.
 *
 * @author EMAC, Inc. <support@emacinc.com>
 *
 * Copyright (C) 2009 EMAC, Inc.
 */
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <fcntl.h>
#include "i2c-dev.h" /* header from lm-sensors project */
#include <stdio.h>
#include <sys/ioctl.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#define FOREVER()  for(;;)

/* i2c prototypes */
int i2c_set_slave(int fd, int addr);
int i2c_write_byte(int fd, __u8 byte);
int i2c_write_cmd(int fd, __u8 reg, __u8 cmd);
int i2c_read_byte(int fd, __u8 *val);
int i2c_read_reg(int fd, __u8 *val, __u8 reg);
int i2c_read_reg2(int fd, __u16 *val, __u8 reg);
int i2c_write_cmd2(int fd, __u8 reg, __u16 cmd);

/* Testing Functions */
int menu(int fd);
void test_set_slave(int fd);
void test_write_byte(int fd);
void test_write_int(int fd);
void test_read_byte(int fd);
void test_read_int(int fd);

/* main */
int main(int argc, __u8 *argv[])
{
	int fd;
	if ((fd = open("/dev/i2c-0", O_RDWR)) < 0) {
		fprintf(stderr, "Unable to open /dev/i2c-0: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	FOREVER() {
		if (menu(fd) == -1)
			break;
	}
	exit(EXIT_SUCCESS);
}

/* i2c read/write functions */

/****************************************************************************
 * NOTE:
 * SMBus is a subset of i2c. Many devices use only this subset of i2c. Linux
 * provides many SMBus functions that can be used to communicate using SMBus.
 * If the i2c device connected uses only the SMBus subset of i2c, it is
 * recommended that these functions are used. See the function inlines in
 * the i2c-dev.h file and the documentation from the Linux kernel included
 * in the docs directory of this project folder. Because nothing is known
 * about the attached device, this test code uses simple i2c read()/write()
 * calls to communicate with the device. Please use the smbus commands if
 * possible.
 ***************************************************************************/

/**
 * function to set the slave address of the i2c device. This only needs to be
 * done to initialize the address and then each time that a new device is
 * addressed. 
 * @note this is the lower 7 bits of the devices slave address
 * @param fd the file descriptor of the open i2c device
 * @param addr the slave address to set
 * @return -1 on error 0 on success
 */
inline int i2c_set_slave(int fd, int addr)
{
	if (ioctl(fd, I2C_SLAVE, addr) < 0) {
		fprintf(stderr, "i2c_set_addr: Error on ioctl: %s\n",
				strerror(errno));
		return -1;
	}
	return 0;
}

/**
 * function to write one byte to the i2c device. This should be changed to
 * use the SMBus functions if possible.
 * @param fd the open file descriptor of the i2c device
 * @param data the byte to write to the device
 * @return -1 on failure 0 on success
 */
inline int i2c_write_byte(int fd, __u8 byte)
{
	if (write(fd, &byte, 1) != 1) {
		fprintf(stderr, "i2c_write_byte: Error on write: %s\n", strerror(errno));
		return -1;
	}
	return 0;
}

/**
 * function to write a byte to a specific register in the device. This
 * should be changed to use the SMBus functions if possible.
 * @param fd the open file descriptor for the i2c device
 * @param reg the register to write to
 * @param cmd the data to write to the device
 * @return -1 on failure 0 on success
 */
inline int i2c_write_cmd(int fd, __u8 reg, __u8 cmd)
{
	__u8 buff[2];

	buff[0] = reg;
	buff[1] = cmd;

	if (write(fd, buff, 2) != 2) {
		fprintf(stderr, "i2c_write_command: Error on write: %s\n",
				strerror(errno));
		return -1;
	}
	return 0;
}

/**
 * function to read a byte from the i2c device. This should be changed
 * to use the SMBus commands if possible.
 * @param fd the open file descriptor for the i2c device
 * @param val a pointer to the location to store the data
 * @return -1 on failure 0 on success
 */
inline int i2c_read_byte(int fd, __u8 *val)
{
	if (read(fd, val, 1) != 1) {
		fprintf(stderr, "i2c_read_byte: Error on read: %s\n", strerror(errno));
		return -1;
	}
	return 0;
}

/**
 * function to read the value at a given register in the i2c device.
 * This should be changed to use SMBus calls if possible. Currently
 * works by writing the register followed by reading the next byte.
 * @param fd the file descriptor of the open i2c device
 * @param val the memory to store data in
 * @param reg the register to read from
 * @return -1 on failure 0 on success
 */
inline int i2c_read_reg(int fd, __u8 *val, __u8 reg)
{
	/* write register */
	if (i2c_write_byte(fd, reg) == -1)
		return -1;
	/* read the value at reg */
	if (i2c_read_byte(fd, val) == -1)
		return -1;

	return 0;
}

/**
 * function to read two bytes from the given register. Should be
 * changed to use SMBus commands if possible.
 * @param fd the open file descriptor for the i2c device
 * @param val the memory location to store data at
 * @param reg the register to read from
 * @return -1 on failure 0 on success
 */
int i2c_read_reg2(int fd, __u16 *val, __u8 reg)
{
	__u8 b;

	/* write register */
	if (i2c_write_byte(fd, reg) == -1)
		return -1;
	/* read byte 1 
	 * (note we could also read 2 bytes at a time -- 16-bit devices may need it) */
	if (i2c_read_byte(fd, &b) == -1)
		return -1;
	/* shift and read byte 2 */
	*val = ((__u16)b) << 8;
	if (i2c_read_byte(fd, &b) == -1)
		return -1;
	/* set value */
	*val += b;

	return 0;
}

/**
 * function to write 2 bytes to a register in the device. This should be
 * changed to use the SMBus commands if possible.
 * @param fd the open file descriptor for the i2c device
 * @param reg the register to write to
 * @param cmd the data to write
 * @return -1 on failure 0 on success
 */
int i2c_write_cmd2(int fd, __u8 reg, __u16 cmd)
{
	__u8 buff[3];
	
	/* write register */
	buff[0] = reg;
	buff[1] = (__u8)(cmd >> 8);
	buff[2] = (__u8)(cmd);

	if (write(fd, buff, 3) != 3) {
		fprintf(stderr, "i2c_write_cmd: error on write: %s\n", strerror(errno));
		return -1;
	}
	return 0;
}

/* Testing interface */

int menu(int fd)
{
	__u8 entry[10];
	static const __u8 *funct_list =
		"\nSelect a function to test:      \n"
		"    a: Set slave address          \n"
		"    b: Write byte to register     \n"
		"    c: Write 2 bytes to register  \n"
		"    d: Read byte from register    \n"
		"    e: Read 2 bytes from register \n"
		"    q: Quit                       \n"
		">> ";

	printf("%s", funct_list);
	while (fgets(entry, 10, stdin) && (entry[0] == '\n'))
		;

	switch (toupper(entry[0]))
	{
	case 'A':
		test_set_slave(fd);
		break;
	case 'B':
		test_write_byte(fd);
		break;
	case 'C':
		test_write_int(fd);
		break;
	case 'D':
		test_read_byte(fd);
		break;
	case 'E':
		test_read_int(fd);
		break;
	case 'Q':
		printf("Quiting...\n");
		close(fd);
		return -1;
		break;
	default:
		printf("Invalid selection (%c), try again.\n", entry[0]);
		break;
	} // end switch

	return 0;
}

void test_set_slave(int fd)
{
	int addr;
	printf("\nEnter the slave address (in hex): ");
	scanf("%X", &addr);
	printf("\nCalling i2c_set_slave(%d, 0x%X)\n", fd, addr);
	if (i2c_set_slave(fd, addr) == -1)
		printf("i2c_set_slave failed\n");
	else
		printf("i2c_set_slave reports success\n");
}

void test_write_byte(int fd)
{
	int cmd;
	int reg;

	printf("\nEnter a byte to write (in hex): ");
	scanf("%X", &cmd);
	printf("Enter a register value to write (in hex): ");
	scanf("%X", &reg);
	printf("\nCalling i2c_write_cmd(%d, 0x%X, 0x%X)\n", fd, reg, cmd);

	if (i2c_write_cmd(fd, (__u8)reg, (__u8)cmd) == -1)
		printf("i2c_write_cmd failed\n");
	else
		printf("i2c_write_cmd reports success\n");
}

void test_write_int(int fd)
{
	int cmd;
	int reg;

	printf("\nEnter a 2 byte number to write (in hex): ");
	scanf("%X", &cmd);
	printf("Enter a register value to write (in hex): ");
	scanf("%X", &reg);
	printf("\nCalling i2c_write_cmd2(%d, 0x%X, 0x%X)\n", fd, reg, cmd);

	if (i2c_write_cmd2(fd, (__u8)reg, (__u16)cmd) == -1)
		printf("i2c_write_cmd2 failed\n");
	else
		printf("i2c_write_cmd2 reports success\n");
}

void test_read_byte(int fd)
{
	int val = 0;
	int reg;

	printf("\nEnter a register value to read from (in hex): ");
	scanf("%X", &reg);
	printf("\nCalling i2c_read_reg(%d, (__u8 *)&val, 0x%X)\n", fd, reg);

	if (i2c_read_reg(fd, (__u8 *)&val, reg) == -1)
		printf("i2c_read_reg failed\n");
	else
		printf("i2c_read_reg reports success\nRead value 0x%X\n", val);
}

void test_read_int(int fd)
{
	int val;
	int reg;

	printf("\nEnter a register value to read from (in hex): ");
	scanf("%X", &reg);
	printf("\nCalling i2c_read_reg2(%d, &val, 0x%X)\n", fd, reg);

	if (i2c_read_reg2(fd, (__u16 *)&val, reg) == -1)
		printf("i2c_read_reg2 failed\n");
	else
		printf("i2c_read_reg2 reports success\nRead value 0x%X\n", val);
}

// EOF
